import os
import psutil
import torch
from accelerate import Accelerator, DistributedType
from dataclasses import dataclass
from typing_extensions import Union

@dataclass
class Monitor:
    """
    Class to set metrics to monitor during training using a tracker (if implemented).

    Args:
        learning_rate (`bool`, *optional*, defaults to `False`):
            Monitor learning rate.
        train_loss (`bool`, *optional*, defaults to `True`):
            Monitor training loss.
        validation_loss (`bool`, *optional*, defaults to `True`):
            Monitor validation loss.
        gpu_utilization (`bool`, *optional*, defaults to `False`):
            Monitor GPU utilization in GB. It only reports GPU from main process (for now).
        cpu_utilization (`bool`, *optional*, defaults to `False`):
            Monitor CPU utilization in GB. It only reports CPU from main process (for now)
    """
    def __init__(self,
                 learning_rate: bool = False,
                 train_loss: bool = True,
                 validation_loss: bool = True,
                 gpu_utilization: bool = False,
                 cpu_utilization: bool = False
    ):
        self.learning_rate = learning_rate
        self.train_loss = train_loss
        self.validation_loss = validation_loss
        self.gpu_utilization = gpu_utilization
        self.cpu_utilization = cpu_utilization
        self.status_dict = None
        self.accelerator = None
        self.train_loss_name = None
        self.validation_loss_name = None

    @classmethod
    def from_config(cls, config: Union[str, dict]):
        """
        Load a monitor configuration from a file or a dictionary.

        Args:
            config (`str` or `dict`):
                Path to a file or dictionary containing kwargs for Monitor constructor. The file can 
                be YAML or JSON.
        """
        assert config is None or isinstance(config, (str, dict)), f"{config} is not of type 'str' or 'dict'."
        if isinstance(config, str):
            import yaml
            config = yaml.safe_load(open(config))
        elif config is None:
            config = {}

        return Monitor(**config)
    
    def _set_extra(self, accelerator: Accelerator, status_dict: dict, train_loss_name: str, validation_loss_name: str):
        self.accelerator = accelerator
        self.status_dict = status_dict
        self.train_loss_name = train_loss_name
        self.validation_loss_name = validation_loss_name

    def log_learning_rate(self):
        if self.learning_rate and self.accelerator.is_main_process:
            self.accelerator.log({"learning_rate": self.status_dict["learning_rate"]}, step=self.status_dict["global_step"]+1)

    def log_train_loss(self):
        if self.train_loss and self.accelerator.is_main_process:
            self.accelerator.log({self.train_loss_name: self.status_dict["train_loss"]}, step=self.status_dict["global_step"]+1)

    def log_validation_loss(self):
        if self.validation_loss and self.accelerator.is_main_process:
            self.accelerator.log({self.validation_loss_name: self.status_dict["validation_loss"]}, step=self.status_dict["eval_global_step"]+1)

    def log_gpu_utilization(self):
        if self.gpu_utilization and self.accelerator.is_main_process:
            if self.accelerator.distributed_type in {
                DistributedType.DEEPSPEED,
                DistributedType.FSDP,
                DistributedType.MULTI_GPU
            }:
                device = torch.device("cuda")
                memory_allocated = torch.cuda.memory_allocated(device)
                memory_reserved = torch.cuda.memory_reserved(device)
                total_memory = (memory_allocated + memory_reserved) / (1024**3)
            
                self.accelerator.log({"GPU_0": total_memory}, step=self.status_dict["global_step"]+1)

    def log_cpu_utilization(self):
        if self.cpu_utilization and self.accelerator.is_main_process:
            process = psutil.Process(os.getpid())
            cpu_mem = process.memory_info().rss / (1024**3)
            self.accelerator.log({"CPU_PROCESS_0": cpu_mem}, step=self.status_dict["global_step"]+1)



from dataclasses import dataclass
from accelerate.utils import LoggerType

@dataclass
class TensorBoard:
    tracker = LoggerType.TENSORBOARD
    init = lambda **kwargs: {"tensorboard": {**kwargs}}

@dataclass
class WandB:
    tracker = LoggerType.WANDB
    init = lambda **kwargs: {"wandb": {**kwargs}}

@dataclass
class CometML:
    tracker = LoggerType.COMETML
    init = lambda **kwargs: {"comet_ml": {**kwargs}}

@dataclass
class Aim:
    tracker = LoggerType.AIM
    init = lambda **kwargs: {"aim": {**kwargs}}

@dataclass
class MLFlow:
    tracker = LoggerType.MLFLOW
    init = lambda **kwargs: {"mlflow": {**kwargs}}

@dataclass
class ClearML:
    tracker = LoggerType.CLEARML
    init = lambda **kwargs: {"clearml": {**kwargs}}

@dataclass
class DVCLive:
    tracker = LoggerType.DVCLIVE
    init = lambda **kwargs: {"dvclive": {**kwargs}}



import torch
from torch.optim import lr_scheduler
from dataclasses import dataclass
from typing_extensions import Optional, Union
from transformers import (
    get_cosine_schedule_with_warmup,
    get_constant_schedule,
    get_constant_schedule_with_warmup,
    get_cosine_with_hard_restarts_schedule_with_warmup,
    get_inverse_sqrt_schedule,
    get_linear_schedule_with_warmup,
    get_polynomial_decay_schedule_with_warmup,
    Adafactor
)

@dataclass
class Optimizer:
    Adam = torch.optim.Adam
    Adadelta = torch.optim.Adadelta
    Adagrad = torch.optim.Adagrad
    Adamax = torch.optim.Adamax
    AdamW = torch.optim.AdamW
    Adafactor = Adafactor
    ASGD = torch.optim.ASGD
    NAdam = torch.optim.NAdam
    RAdam = torch.optim.RAdam
    RMSprop = torch.optim.RMSprop
    Rprop = torch.optim.Rprop
    SGD = torch.optim.SGD
    SparseAdam = torch.optim.SparseAdam

@dataclass
class Scheduler:
    StepLR = lr_scheduler.StepLR
    LinearLR = lr_scheduler.LinearLR
    ExponentialLR = lr_scheduler.ExponentialLR
    CosineAnnealingLR = lr_scheduler.CosineAnnealingLR
    CyclicLR = lr_scheduler.CyclicLR
    OneCycleLR = lr_scheduler.OneCycleLR
    CosineAnnealingWarmRestarts = lr_scheduler.CosineAnnealingWarmRestarts
    CosineWithWarmup = get_cosine_schedule_with_warmup
    Constant = get_constant_schedule
    ConstantWithWarmup = get_constant_schedule_with_warmup
    CosineWithHardRestartsWithWarmup = get_cosine_with_hard_restarts_schedule_with_warmup
    InverseSQRT = get_inverse_sqrt_schedule
    LinearWithWarmup = get_linear_schedule_with_warmup
    PolynomialDecayWithWarmup = get_polynomial_decay_schedule_with_warmup

@dataclass
class HyperParameters:
    """
    Class to set hyperparameters for training.

    Args:
        epochs (`int`, *optional*, defaults to `1`):
            Number of epochs (how many times we run the model over the dataset).
        batch_size (`int` or `tuple`, *optional*, defaults to `1`):
            Batch size (how many samples are passed to the model at the same time). This can also be a 
            `tuple`, the first element indicating batch size during training, and the second element 
            indicating batch size during evaluation.

            NOTE: This is not effective batch size. Effective batch size will be calculated multiplicating 
            this value by the number of processes.
        optim (`str` or `Optimizer`, *optional*, defaults to `SGD`):
            Optimization algorithm. See documentation to check the available ones.
        optim_kwargs (`dict`, *optional*, defaults to `None`):
            Specific optimizer keyword arguments.
        scheduler (`str` or `Scheduler`, *optional*, defaults to `None`):
            Learning rate scheduler to implement.
        scheduler_kwargs (`dict`, *optional*, defaults to `None`):
            Specific scheduler keyword arguments.
    """
    epochs: int = 1
    batch_size: Union[int, tuple[int]] = 1
    optim: Union[str, Optimizer] = "SGD"
    optim_kwargs: Optional[dict] = None
    scheduler: Optional[Union[str, Scheduler]] = None
    scheduler_kwargs: Optional[dict] = None

    def to_dict(self) -> dict:
        optim = self.optim if not isinstance(self.optim, str) else getattr(Optimizer, self.optim, None)
        assert optim is not None, f"{optim} is not a valid optimizer."
        scheduler = self.scheduler if not isinstance(self.scheduler, str) else getattr(Scheduler, self.scheduler, "INVALID")
        assert scheduler != "INVALID", f"{scheduler} is not a valid scheduler."

        optim_kwargs = self.optim_kwargs if self.optim_kwargs is not None else {}
        schlr_kwargs = self.scheduler_kwargs if self.scheduler_kwargs is not None else {}

        optim_name = str(optim).removesuffix("'>").split(".")[-1] if not isinstance(optim, str) else optim
        d = {
            "hps": {
                "epochs": self.epochs,
                "batch_size": self.batch_size,
                "optim": {"type": optim_name, **optim_kwargs}
            }
        }

        if self.scheduler is not None:
            scheduler_name = str(scheduler).removesuffix("'>").split(".")[-1] if not isinstance(scheduler, str) else scheduler
            d["hps"]["scheduler"] = {"type": scheduler_name, **schlr_kwargs}

        return d

